/*
 * packetmodeller/packet_modeller.c
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "init.h"
#include "message.h"
#include "model.h"
#include "netdevice.h"
#include "packet_modeller.h"
#include "pm_print.h"


/** Print usage details */
void
pm_usage()
{
    printf("\nUsage: %s MODELIP:MODELPORT LOCALPORT\n", exec_name);
    printf("\t\t\t[-d][-h][-i INTERFACE][-n]\n");
    printf("Try `%s -h' for more details\n", exec_name);

    exit(EXIT_FAILURE);
}

/** Print help */
void
pm_help()
{
    printf("\nUsage: %s MODELIP:MODELPORT LOCALPORT\n", exec_name);
    printf("\t\t\t[-d][-h][-i INTERFACE][-n]\n");
    printf("\nTunnels traffic from a local network to another %s", exec_name);
    printf("instance\nor a wireless emulation model\n");
    printf("\nParameters:\n");
    printf("MODELIP:PORT\tIP Address and port number of tunnel end point\n");
    printf("LOCALPORT\tLocal port to listen for connections\n");
    printf("-d\t\tEnable debug output\n");
    printf("-h\t\tShow this help message\n");
    printf("-i INTERFACE\tTunnel interface (default: %s)\n", node_pif);
    printf("-n\t\tEnable node-to-node communication (only for testing!)\n");

    exit(EXIT_SUCCESS);
}

/** Stub for print_debug prototype */
int
ignore_debug(const char* text, ...)
{
    return 0;
}

/** Print debug message with variable argument list */
int
print_debug(const char *format, ...)
{
    va_list arg;

    va_start(arg, format);
    pm_debug_args(format, arg);
    va_end(arg);

    return 0;
}

/** Stub for print_debug_args prototype */
void
ignore_debug_args(const char *format, va_list arg)
{
}

/** Print debug message with given argument list */
void
print_debug_args(const char *format, va_list arg)
{
    FILE *stdout = fopen("/dev/stdout", "a");
    fprintf(stdout, "DEBUG: ");
    vfprintf(stdout, format, arg);
    fclose(stdout);
}

/** Print Error message */
void
pm_error(const char *format, ...)
{
    va_list arg;

    FILE *stderr = fopen("/dev/stderr", "a");
    fprintf(stderr, "Error: ");
    va_start(arg, format);
    vfprintf(stderr, format, arg);
    va_end(arg);
    fclose(stderr);
}

/** Send message to syslog */
void
pm_log(const char *format, ...)
{
    va_list arg;

    openlog(exec_name, LOG_PID, LOG_SYSLOG);
    va_start(arg, format);
    vsyslog(LOG_INFO, format, arg);
    pm_debug_args(format, arg);
    va_end(arg);
    closelog();
}

/** Parse command line options and checks it for validity */
void
pars_opts(int argc, char *argv[])
{
    int opt;
    int debug = 0;
    int req_args = 3;
    int add_args = 0; // required because getopts changes the order of arguments
    direct = 0;
    strcpy(node_pif, "eth0");

    // IMPORTANT: Don't use pm_log or pm_debug before here!
    pm_debug = ignore_debug;
    pm_debug_args = ignore_debug_args;

    while ((opt = getopt(argc, argv, "dnhi:")) != -1)
    {
        switch (opt)
        {
            case 'h':
                pm_help();
                break;
            case 'i':
                add_args += 2;
                break;
            case 'd':
                add_args++;
                debug = 1;
                break;
            case 'n':
                add_args++;
                break;
        }
    }
    optind = 0; // reset getopt execution

    if (argc < (req_args + add_args))
    {
        pm_error("Missing arguments\n");
        pm_usage();
    }

    if (debug == 1)
    {
        pm_debug = print_debug;
        pm_debug_args = print_debug_args;
    }

    // read model address
    if (validate_model_addr(&model_addr, argv[1 + add_args]) == -1)
        pm_usage();

    // read local port
    if ((node_port = validate_port(argv[2 + add_args])) == 0)
        pm_usage();

    while ((opt = getopt(argc, argv, "dni:")) != -1)
    {
        switch (opt)
        {
            case 'n':
                direct = 1;
                break;
            case 'i':
                if (validate_dev_name(optarg))
                    pm_usage();
                break;
        }
    }
    pm_log("Starting PacketModeller (v0.1.0)\n");
    if (debug == 1)
        pm_log("Debugging enabled!\n");
    pm_log("Using physical interface '%s'\n", node_pif);
}

/** Set wireless information to Ethernet frames and send to model */
void
process_data_from_tap(struct wlan_msg *message, const int ether_length)
{
    (*message).header.type = WLAN_MSG;
    memmove(&(*message).header.id, node_id, sizeof(host_id));
    (*message).pkt_length = ether_length;
    debug_msg((char *) message);

    /* send data to model */
    if (sendto(net_socket, message, WLAN_PREAMBLE_SIZE + ether_length, 0, \
        (struct sockaddr *) &model_addr, sizeof(struct sockaddr_in)) < 0)
    {
        pm_error("Sending data failed (%s)\n", strerror(errno));
    }
    pm_debug("process_data_from_tap(): [net_socket] %i Bytes sent (%s)\n", \
        WLAN_PREAMBLE_SIZE + ether_length, strerror(errno));
}

/** Process packets received from WlanModel and inject to local network */
void
process_data_from_net(char *buf, const int message_length)
{
    int payload_length;
    raw_packet *ether_pkt;

    // received message from model
    if (isWlanMessage((struct msg_header *) buf))
    {
        struct wlan_msg *message = (struct wlan_msg *) buf;
        ether_pkt = (raw_packet *) &((*message).wlan_pkt);
        payload_length = message_length - WLAN_PREAMBLE_SIZE;
        debug_msg(buf);
    } else {
        pm_debug("process_data_from_net(): Unknown message received\n");
        return;
    }

    // send to TAP interface
    if (write(tap_fd, ether_pkt, payload_length) < 0)
    {
        pm_error("Couldn't locally deliver packet (%s)\n", strerror(errno));
    }
    pm_debug("process_data_from_net(): [tap_fd] %i Bytes sent (%s)\n", \
        payload_length, strerror(errno));
}

/** Listen to TUN and network interfaces for "wireless" packets */
void
pm_daemon()
{
    fd_set fds;
    int recv_bytes;
    register int ret;
    // wlan message is biggest message possible
    char buffer[WLAN_MSG_SIZE];

    fm = max(tap_fd, net_socket) + 1;
    pm_debug("pm_daemon(): running main loop\n");
    for (;;)
    {
        FD_ZERO(&fds);
        FD_SET(tap_fd, &fds);
        FD_SET(net_socket, &fds);
        if ((ret = select(fm, &fds, NULL, NULL, NULL)) < 0)
        {
            pm_error("Couldn't get active file descriptor (%s)\n", \
                strerror(errno));
        } else {
            // check tap device for incoming traffic
            if (FD_ISSET(tap_fd, &fds))
            {
                struct wlan_msg *message = (struct wlan_msg *) buffer;
                recv_bytes = read(tap_fd, &(*message).wlan_pkt, MAX_FRAME_SIZE);
                pm_debug("pm_daemon(): [tap_fd] %i Bytes read (%s)\n", \
                    recv_bytes, strerror(errno));
                if (recv_bytes < 0)
                {
                    pm_error("Reading data from TAP failed (%s)\n", \
                        strerror(errno));
                } else {
                    process_data_from_tap(message, recv_bytes);
                }
            }
            // check network device for incoming traffic
            if (FD_ISSET(net_socket, &fds))
            {
                recv_bytes = recvfrom(net_socket, buffer, MAX_MSG_SIZE, 0, \
                    NULL, NULL);
                pm_debug("pm_daemon(): [net_socket] %i Bytes read (%s)\n", \
                    recv_bytes, strerror(errno));
                if (recv_bytes < 0)
                {
                    pm_error("Reading data from network failed (%s)\n", \
                        strerror(errno));
                } else {
                    process_data_from_net(buffer, recv_bytes);
                }
            }
        }
    }
}

/** MAIN */
int
main(int argc, char *argv[])
{
    pthread_t model_notifier;

    if (strrchr(argv[0], '/') == NULL)
    {
        exec_name = argv[0];
    } else {
        exec_name = strrchr(argv[0], '/') + 1;
    }

    // check parameters
    pars_opts(argc, argv);

    // initialize node values
    if (pm_init() != 0)
        exit(EXIT_FAILURE);

    // register node at WlanModel
    if (direct != 1)
    {
        if (pm_register() != 0)
            exit(EXIT_FAILURE);
    } else {
        pm_log("Argument '-n' found: Not registering host!\n");
    }

    // create thread for updating model at vif configuration changes
    if (pthread_create(&model_notifier, NULL, vif_handler, NULL) < 0)
    {
        pm_error("Couldn't create thread model_notifier (%s)\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    // run packet redirecting daemon
    pm_daemon();

    exit(EXIT_SUCCESS);
}
