/*
 * vifd/libvif.h
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/**
 * Small library for setting and accessing virtual interface wireless
 * parameters
 */

#ifndef __LIBVIF_H
#define __LIBVIF_H

/** INCLUDES ******************************************************************/
#include <errno.h>
#include <iwlib.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <linux/wireless.h>
#include <net/if.h>
#include <sys/socket.h>
#include <sys/types.h>
#include "vifd_ipc.h"


/** VIF DATATYPES *************************************************************/
typedef char if_name [IF_NAMESIZE + 1];

/** IOCTL handler */
typedef struct ioctl_handler
{
    int ioctl_num;
    int (*iw_handler)(struct iwreq * request, struct wireless_info * vif_config);
    __u8 data_flag;
} ioctl_handler;


/** VIF VALUES ****************************************************************/
// Define program pathes (TODO: replace from configure)
#define TUNCTL_PATH     "/usr/sbin/tunctl"
#define IP_PATH         "/bin/ip"

/** Maximal number of virtual interfaces */
#define MAX_NUM_VIF     2

/** The frequency of each channel in MHz */
static const long
channel_frequency[] = {
    2412, 2417, 2422, 2427, 2432, 2437, 2442,
    2447, 2452, 2457, 2462, 2467, 2472, 2484
};


/** VIF VARIABLES *************************************************************/

/** Number of virtual wireless interfaces */
int
num_vif;

/** List of virtual wireless interface configurations */
struct vif_info *
vif_list[MAX_NUM_VIF];

/** Message queue identifiers */
extern int iw_msgq_id;
extern int pm_msgq_id;

/** Message queue identifiers */
extern int iw_shm_id;
extern int pm_shm_id;


/** PUBLIC FUNCTIONS **********************************************************/

/**
 * Create wireless configuration with default settings
 * @param - device name
 * @param - device index (< MAX_NUM_VIF)
 */
void
create_vif_config(const char *, int index);

/**
 * perform ioctl on virtual wireless device
 * @param - IPC message with iwreq
 */
void
handle_iwreq(struct iwreq_msg *);

/**
 * sets wireless information for virtual device
 * @param prop - property name
 * @param req_vif - vif configuration
 */
int
set_vif_prop(int prop, struct iwreq *req_vif);


/** PRIVATE FUNCTIONS (DON'T CALL FROM OUTSIDE) *******************************/

int
get_name(struct iwreq *, struct wireless_info *);

int
get_nwid(struct iwreq *, struct wireless_info *);

int
set_freq(struct iwreq *, struct wireless_info *);

int
get_freq(struct iwreq *, struct wireless_info *);

int
set_mode(struct iwreq *, struct wireless_info *);

int
get_mode(struct iwreq *, struct wireless_info *);

int
set_sens(struct iwreq *, struct wireless_info *);

int
get_sens(struct iwreq *, struct wireless_info *);

int
get_range(struct iwreq *, struct wireless_info *);

int
get_priv(struct iwreq *, struct wireless_info *);

int
get_apaddr(struct iwreq *, struct wireless_info *);

int
set_essid(struct iwreq *, struct wireless_info *);

int
get_essid(struct iwreq *, struct wireless_info *);

int
get_nick(struct iwreq *, struct wireless_info *);

int
set_bitrate(struct iwreq *, struct wireless_info *);

int
get_bitrate(struct iwreq *, struct wireless_info *);

int
set_rts(struct iwreq *, struct wireless_info *);

int
get_rts(struct iwreq *, struct wireless_info *);

int
set_frag(struct iwreq *, struct wireless_info *);

int
get_frag(struct iwreq *, struct wireless_info *);

int
set_txpower(struct iwreq *, struct wireless_info *);

int
get_txpower(struct iwreq *, struct wireless_info *);

int
set_retry(struct iwreq *, struct wireless_info *);

int
get_retry(struct iwreq *, struct wireless_info *);

int
set_encoding(struct iwreq *, struct wireless_info *);

int
get_encoding(struct iwreq *, struct wireless_info *);

int
get_power(struct iwreq *, struct wireless_info *);


/** IMPLEMENTED IOCTLS ********************************************************/

/** List of supported standard ioctls (see wireless.h for details) */
static const struct ioctl_handler ioctl_handler_list[] = {
    { SIOCGIWNAME, get_name, 0 },
    { SIOCGIWNWID, get_nwid, 0 },
    { SIOCSIWFREQ, set_freq, 0 },
    { SIOCGIWFREQ, get_freq, 0 },
    { SIOCSIWMODE, set_mode, 0 },
    { SIOCGIWMODE, get_mode, 0 },
    { SIOCSIWSENS, set_sens, 0 },
    { SIOCGIWSENS, get_sens, 0 },
    { SIOCGIWRANGE, get_range, IWREQ_DATA },
    { SIOCGIWPRIV, get_priv, IWREQ_DATA },
    { SIOCGIWAP, get_apaddr, 0 },
    { SIOCSIWESSID, set_essid, IWREQ_DATA },
    { SIOCGIWESSID, get_essid, IWREQ_DATA },
    { SIOCGIWNICKN, get_nick, IWREQ_DATA },
    { SIOCSIWRATE, set_bitrate, 0 },
    { SIOCGIWRATE, get_bitrate, 0 },
    { SIOCSIWRTS, set_rts, 0 },
    { SIOCGIWRTS, get_rts, 0 },
    { SIOCSIWFRAG, set_frag, 0 },
    { SIOCGIWFRAG, get_frag, 0 },
    { SIOCSIWTXPOW, set_txpower, 0 },
    { SIOCGIWTXPOW, get_txpower, 0 },
    { SIOCSIWRETRY, set_retry, 0 },
    { SIOCGIWRETRY, get_retry, 0 },
    { SIOCSIWENCODE, set_encoding, IWREQ_DATA },
    { SIOCGIWENCODE, get_encoding, IWREQ_DATA },
    { SIOCGIWPOWER, get_power, 0 }
};

/** List of ioctls returned via SIOCGIWPRIV (for iwpriv compatibility) */
static const struct iw_priv_args vif_private_args[] = {
    { SIOCSIWMODE, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "mode" }
};

#endif
