/*
 * wlanmodel/message.h
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef __MESSAGE_H
#define __MESSAGE_H

/** INCLUDES ******************************************************************/
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "pm_types.h"


/** MESSAGE PARAMETER DEFINITIONS *********************************************/

// maximal size of an Ethernet frame (MTU 1500);
#define MAX_FRAME_SIZE          (1500 + sizeof(struct ether_header))

#define WLAN_MSG                0
#define WLAN_PREAMBLE_SIZE      (sizeof(struct msg_header) + \
                                sizeof(uint16_t))
#define WLAN_MSG_SIZE           WLAN_PREAMBLE_SIZE + MAX_FRAME_SIZE

#define CONFIG_MSG              1
#define CONFIG_MSG_SIZE         (sizeof(struct msg_header) + \
                                sizeof(uint8_t) + \
                                sizeof(uint16_t) + \
                                sizeof(union vif_param))

#define REGISTER_MSG            2
#define REGISTER_MSG_SIZE       (sizeof(struct msg_header) + \
                                sizeof(uint16_t) + \
                                HOST_NAME_MAX * sizeof(char) + \
                                sizeof(struct in_addr) + \
                                sizeof(uint16_t) + \
                                sizeof(model_vif_info))

#define ACK_MSG                 3
#define ACK_MSG_SIZE            (sizeof(struct msg_header) + \
                                sizeof(uint8_t))

// maximal size of a node <-> model message
#define MAX_MSG_SIZE            max(WLAN_MSG_SIZE, REGISTER_MSG_SIZE)


/** MESSAGE TYPES DEFINITIONS *************************************************/

typedef char raw_packet [MAX_FRAME_SIZE];

/**
 *  Message header
 *  @var type: message type
 *  @var id: sender identification
 */
struct msg_header {
    unsigned char type;
    host_id id;
};

/**
 *  Wlan message (encapsulated wireless packet)
 *  @var pkt_length: length of Ethernet packet
 *  @var wlan_pkt: raw Ethernet packet
 */
struct wlan_msg {
    struct msg_header header;
    uint16_t pkt_length;
    raw_packet wlan_pkt;
};

/**
 *  Configuration message
 *  @var type: parameter type
 */
struct config_msg {
    struct msg_header header;
    uint8_t type;
    uint16_t property;
    union vif_param value;
};

/**
 *  Register message
 *  @var msg_id: message identification
 *  @var name: node hostname
 *  @var ip: node IP address
 *  @var port: node listen port
 *  @var vif_info: information about virtual interface
 */
struct register_msg {
    struct msg_header header;
    uint16_t msg_id;
    char name[HOST_NAME_MAX];
    struct in_addr ip;
    uint16_t port;
    model_vif_info vif_info;
};

/**
 *  Acknowledge message
 *  @var msg_id: message identification (same as register msg_id)
 */
struct ack_msg {
    msg_header header;
    unsigned char msg_id;
};

/**
 *  Message debug functions
 */
void
debug_msg(const char *message);

/**
 *  Ignore debug stub
 */
int ignore_dbg(const char* text, ...);

#endif
